﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(_IW_GAME_ACTOR_H_)
#define _IW_GAME_ACTOR_H_

#include "IwGx.h"
#include "Iw2D.h"
#include "IwGeom.h"

#include "IwGameUtil.h"
#include "IwGameSprite.h"
#include "IwGameAnim.h"
#include "IwGameString.h"
#include "IwGameXoml.h"
#include "IwGameBox2d.h"
#include "IwGameShapes.h"
#include "IwGameActions.h"

class CIwGameScene;

//
//
//
//
// CIwGameActor - A basic game actor. We derive different types of game objects from this type and add them to a IwGameScene to be processed and rendered
//
//
//
//
class CIwGameActor : public IIwGameXomlResource, public IIwGameAnimTarget
{
public:

protected:
	// Properties
	CIwGameScene*		Scene;					// Scene that actor lives in
	bool				Used;					// Used is used when actors pooled to reduce memory fragmentation
	int					Type;					// Type of Actor (use to distinguish beteeen different actor types)
	CIwFVec2			OriginalPosition;		// Original position of actor in the scene (when actor was first spawned)
	CIwFVec2			Position;				// Current position of actor in the scene
	float				Depth;					// Parallax depth
	CIwFVec2			Origin;					// Origin of actor
	CIwFVec2			Velocity;				// Current velocity of actor
	CIwFVec2			VelocityDamping;		// Dampens the velocity (Values less than 1.0f will reduce velocity over time, whilst values larger than 1.0f will increase velocity over time)
	float				OriginalAngle;			// Original angle in scene (when first spawned)
	float				Angle;					// Orientation in scene (degrees)
	float				AngularVelocity;		// Angular velocity
	float				AngularVelocityDamping;	// Angular velocity damping
	float				ScaleX, ScaleY;			// Scale
	CIwColour			Colour;					// Colour
	int					Layer;					// Depth layer
	bool				IsActive;				// Active state of actor
	bool				IsVisible;				// Visible state of actor
	bool				IsCollidable;			// Collidable state of actor
	bool				IsTappable;				// Tappable state of actor
	CIwGameSprite*		Visual;					// Visual element that represents the actor
	CIwGameAnimTimeline* Timeline;				// Timeline, controls animation of the actor
	CIwGameXomlBindings* Bindings;				// Manages any attached bindings
	int					CollisionSize;			// Size of collision area
	CIwRect				CollisionRect;			// Spherical collision size
	float				PreviousAngle;			// Previous updates angle
	CIwFVec2			PreviousPosition;		// Previous updates position
	bool				WrapPosition;			// If true then if position exits the extents of the scene actor will wrap back around to the opposite side of the scene
	CIwGameActor*		LinkedTo;				// Actor that this actor links to. Linked actors will utilise their target links colour, visibility and transform etc..
	CIwGameActor*		Target;					// An actor target
	bool				Destroyed;				// An actor is marked asd destroyed when it has been marked for deletion
#if defined(IW_GAME_BOX2D)
	CIwGameBox2dBody*	Box2dBody;				// Box2D physics body
#endif // IW_GAME_BOX2D_H_
	CIwGameEventManager* EventsManager;			// List of events that the actor handles
public:
	void				setUsed(bool in_use)					{ Used = in_use; }
	bool				isUsed() const							{ return Used; }
	void				setScene(CIwGameScene* scene)			{ Scene = scene; }
	CIwGameScene*		getScene()								{ return Scene; }
	void				setType(int type)						{ Type = type; }
	int					getType() const							{ return Type; }
	void				setOriginalPosition(float x, float y)	{ OriginalPosition.x = x; OriginalPosition.y = y; }
	CIwFVec2			getOriginalPosition()					{ return OriginalPosition; }
	void				setPosition(float x, float y);
	CIwFVec2			getPosition()							{ return Position; }
	void				setDepth(float depth)					{ Depth = depth; }
	float				getDepth() const						{ return Depth; }
	void				setOrigin(float x, float y)				{ Origin.x = x; Origin.y = y; }
	CIwFVec2			getOrigin()								{ return Origin; }
	void				setOriginalAngle(float angle)			{ OriginalAngle = angle; }
	float				getOriginalAngle()						{ return OriginalAngle; }
	void				setAngle(float angle);
	float				getAngle() const						{ return Angle; }
	void				setVelocity(float x, float y);
	CIwFVec2			getVelocity() const						{ return Velocity; }
	void				setVelocityDamping(float x, float y);	// When using Box2d, the y value is ignored
	void				setVelocityDamping(float damping);
	void				setAngularVelocity(float velocity);
	float				getAngularVelocity() const				{ return AngularVelocity; }
	void				setAngularVelocityDamping(float damping);
	void				setScale(float scale)					{ ScaleX = ScaleY = scale; }
	float				getScale() const						{ return ScaleX; }
	void				setScaleX(float scale)					{ ScaleX = scale; }
	float				getScaleX() const						{ return ScaleX; }
	void				setScaleY(float scale)					{ ScaleY = scale; }
	float				getScaleY() const						{ return ScaleY; }
	void				setColour(uint8 r, uint8 g, uint8 b, uint8 a)	{ Colour.r = r; Colour.g = g; Colour.b = b; Colour.a = a; }
	void				setColour(CIwColour& colour)			{ Colour = colour; }
	CIwColour			getColour() const						{ return Colour; }
	void				setLayer(int layer)						{ Layer = layer; }
	int					getLayer() const						{ return Layer; }
	void				setActive(bool active);
	bool				isActive() const						{ return IsActive; }
	void				setVisible(bool visible)				{ IsVisible = visible; }
	bool				isVisible() const						{ return IsVisible; }
	void				setCollidable(bool collidable);
	bool				isCollidable() const					{ return IsCollidable; }
	void				setTappable(bool tappable)				{ IsTappable = tappable; }
	bool				isTappable() const						{ return IsTappable; }
	void				setVisual(CIwGameSprite* visual)		{ Visual = visual; }
	CIwGameSprite*		getVisual()								{ return Visual; }
	void				removeVisual();
	void				setTimeline(CIwGameAnimTimeline* timeline)	{ Timeline = timeline; }
	CIwGameAnimTimeline* getTimeline()							{ return Timeline; }
	void				setBindings(CIwGameXomlBindings* bindings)	{ Bindings = bindings; }
	CIwGameXomlBindings* getBindings()							{ return Bindings; }
	void				setCollisionRect(CIwRect& rect);
	void				setCollisionSize(int size);
	CIwRect				getCollisionRect() const				{ return CollisionRect; }
	int					getCollisionSize() const				{ return CollisionSize; }
	void				setPreviousPosition(float x, float y)	{ PreviousPosition.x = x; PreviousPosition.y = y; }
	CIwFVec2			getPreviousPosition() const				{ return PreviousPosition; }
	void				setPreviousAngle(float angle)			{ PreviousAngle = angle; }
	float				getPreviousAngle() const				{ return PreviousAngle; }
	void				setWrapPosition(bool enable)			{ WrapPosition = enable; }
	bool				getWrapPosition() const					{ return WrapPosition; }
	void				setLinkedTo(CIwGameActor* actor)		{ LinkedTo = actor; }
	CIwGameActor*		getLinkedTo()							{ return LinkedTo; }
	void				setTarget(CIwGameActor* target)			{ Target = target; }
	CIwGameActor*		getTarget()								{ return Target; }
	void				setDestroyed(bool destroyed)			{ Destroyed = destroyed; }
	bool				isDestroyed() const						{ return Destroyed; }
#if defined(IW_GAME_BOX2D)
	void				setBox2dBody(CIwGameBox2dBody* body)	{ Box2dBody = body; Box2dBody->setUserData(this); }
	CIwGameBox2dBody*	getBox2dBody()							{ return Box2dBody; }
#endif // IW_GAME_BOX2D_H_
	CIwGameEventManager* getEventsManager()						{ return EventsManager; }
	// Properties end
protected:
	virtual void		UpdateBindings();			// Update data bindings

public:
	CIwGameActor() : IIwGameXomlResource(), Used(false), Scene(NULL), Visual(NULL), Timeline(NULL), EventsManager(NULL), LinkedTo(NULL), Bindings(NULL), Destroyed(false)
#if defined(IW_GAME_BOX2D)
	, Box2dBody(NULL)
#endif // IW_GAME_BOX2D_H_
	{
		setClassType("actor");
		Reset();
		EventsManager = new CIwGameEventManager();
	}
	virtual				~CIwGameActor();

	// Initialise the actor
	virtual void		Init();
	
	// Reset the actor (used by memory pooling systems to save actor re-allocation, usually called after re-allocation to reset the object to a default state)
	virtual void		Reset();

	// Update the actor
	virtual bool		Update(float dt);

	// Update the visual that represents this actor on screen
	virtual bool		UpdateVisual();

	// Actors are responsible for carrying out there own collision checks. Called after all actor updates to check and resolve any collisions
	virtual void		ResolveCollisions() {}

	// NotifyCollision is called by another actor when it collides with this actor
	virtual	void		NotifyCollision(CIwGameActor* other) {}

	// Hit test
	virtual CIwGameActor* HitTest(int x, int y);

	// Checks to see if another actor is colliding with this actor
	bool				CheckCollision(CIwGameActor* other);

	// Event handlers
	virtual void		ProcessEventActions(unsigned int event_name);
	virtual void		OnTapped(int x, int y);
	virtual void		OnBeginTouch(int x, int y);
	virtual void		OnEndTouch(int x, int y);
	virtual void		OnCreate();
	virtual void		OnDestroy();

	// Implementation of IIwGameXomlResource interface
	bool				LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node);

	// Implementation of IIwGameAnimTarget interface
	bool				UpdateFromAnimation(CIwGameAnimInstance *animation);

	// Utility
	void				RemoveBody();


};

//
//
//
//
// CIwGameActorImage - An image based game actor. Acts as an helper class for creating image sprite based actors
//
//
//
//
class CIwGameActorImage : public CIwGameActor
{
public:
	// Properties
protected:
	bool				InitWithImage(CIwGameImage* image, int width, int height);
	void				UpdateBindings();
public:
	// Properties end
	CIwGameActorImage() : CIwGameActor()	{ setActualClassType("actorimage"); }
	virtual ~CIwGameActorImage()			{}

	virtual bool		Init(CIwGameImage* image, int width, int height);
	virtual bool		Init(CIwGameImage* image, CIwRect* src_rect);
	virtual bool		Init(CIwGameImage* image, int x, int y, int w, int h);
	bool				UpdateVisual();

	void				SetSrcDest(int x, int y, int w, int h);
	void				SetSrcRect(int x, int y, int w, int h);
	void				SetDestSize(int w, int h);
	void				SetImage(CIwGameImage* image);

	// Implementation of IIwGameXomlClass interface
	bool				LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node);

	// Implementation of IIwGameAnimTarget interface
	bool				UpdateFromAnimation(CIwGameAnimInstance *animation);
};

//
//
//
//
// CIwGameActorText - A textbased game actor. Acts as an helper class for creating text sprite based actors
//
//
//
//
class CIwGameActorText : public CIwGameActor
{
public:
	// Properties
protected:
public:
	void				setText(const char* text);
	void				setRect(CIwRect& rect);
	void				setFlags(IwGxFontFlags flags);
	void				setFont(CIwGameFont* font);
	void				setAlignH(IwGxFontAlignHor align);
	void				setAlignV(IwGxFontAlignVer align);
	// Properties end
protected:
	void				UpdateBindings();

public:
	CIwGameActorText() : CIwGameActor()	{ setActualClassType("actortext"); IsTappable = false; }
	virtual ~CIwGameActorText()			{}

	virtual bool		Init(CIwGameFont* font);
	virtual bool		UpdateVisual();

	// Implementation of IIwGameXomlClass interface
	bool				LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node);

	// Implementation of IIwGameAnimTarget interface
	bool				UpdateFromAnimation(CIwGameAnimInstance *animation);
};

//
//  CIwGameActorTextCreator - Creates an instance of a text based text actor object
//
class CIwGameActorTextCreator : public IIwGameXomlClassCreator
{
public:
	CIwGameActorTextCreator()
	{
		setClassName("actortext");
	}
	IIwGameXomlResource*	CreateInstance(IIwGameXomlResource* parent)	{ return new CIwGameActorText(); }
};

//
//
//
//
//  CIwGameXomlAction_Actors - A collection of actor actions
//
//
//
//
class CIwGameXomlAction_Actors : public IIwGameXomlAction
{
public:
	enum ActionType
	{
		Action_HideActor		= 0, 
		Action_ShowActor		= 1, 
		Action_ActivateActor	= 2, 
		Action_DeactivateActor	= 3, 
		Action_KillActor		= 4, 
		Action_Max				= 5, 
	};
protected:
	ActionType	Type;
	CIwGameXomlAction_Actors() {}
public:
	CIwGameXomlAction_Actors(ActionType type)
	{
		Type = type;
		switch (Type)
		{
		case Action_HideActor:
			setActionName("hideactor");
			break;
		case Action_ShowActor:
			setActionName("showactor");
			break;
		case Action_ActivateActor:
			setActionName("activateactor");
			break;
		case Action_DeactivateActor:
			setActionName("deactivateactor");
			break;
		case Action_KillActor:
			setActionName("killactor");
			break;
		}
	}
	void Execute(IIwGameXomlResource* source, CIwGameAction* action);
};


#endif // _IW_GAME_ACTOR_H_
