// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#if !defined(IW_GAME_SCENE_H_)
#define IW_GAME_SCENE_H_

#include "s3e.h"
#include "IwGx.h"
#include "IwGeom.h"

#include "IwGameSprite.h"
#include "IwGameAnim.h"
#include "IwGameImage.h"
#include "IwGameUtil.h"
#include "IwGameString.h"
#include "IwGameXoml.h"
#include "IwGameResources.h"
#include "IwGameBox2d.h"
#include "IwGameActions.h"

class CIwGameActor;
class CIwGameCamera;

//
//
//
//
//  CIwGameScene - A scene contains, handles and tracks a group of actors
//
//
//
//
class CIwGameScene : public IIwGameXomlResource, public IIwGameAnimTarget
{
public:
	// Public access to actor iteration
	typedef CIwList<CIwGameActor*>::iterator _Iterator;
	_Iterator					begin() { return Actors.begin(); }
	_Iterator					end() { return Actors.end(); }

	// Properties
protected:
	CIwGameSpriteManager*		SpriteManager;		// Manages sprites for the whole scene
	CIwGameXomlResourceManager*	ResourceManager;	// Manages all types of resources
	CIwGameXomlVariableManager*	VariableManager;	// Manages XOML variables
	CIwGameXomlBindings*		Bindings;			// Manages any attached bindings
	CIwGame*					Parent;				// Parent game controller
	int							Type;				// Type of scene
	CIwVec2						ScreenSize;			// Native screen size
	CIwVec2						VirtualSize;		// The virtual size is not the actual size of the scene. but a static pretend size that we can use to render to without having to cater for different sized displays
	CIwMat2D					VirtualTransform;	// Virtual transform is used to scale, translate and rotate scene to fit different display sizes and orientations
	CIwMat2D					Transform;			// Scene transform
	CIwList<CIwGameActor*>		Actors;				// Collection of scene actors
	CIwRect						Extents;			// Extents of scenes world
	CIwGameCamera*				Camera;				// Current camera
	CIwGameActor**				Collidables;		// List of collidable objects built this frame
	bool						AllowSuspend;		// If true then this processing of this scene will be suspended when the scene manager changes away from this scene
	CIwRect						ClippingArea;		// A rectangular area of the virtual screen that this scene will be clipped to
	bool						IsActive;			// Active state of scene
	bool						IsVisible;			// Visible state of scene
	bool						AllowFocus;			// If true this scene can receive input events if it is not the current scene
	CIwColour					Colour;				// Colour
	CIwGameAnimTimeline*		Timeline;			// Timeline, controls animation of the scene
	int							RunCount;			// NUmber of times the scene has been updated
	bool						Destroyed;			// A scene is marked asd destroyed when it has been marked for deletion
#if defined(IW_GAME_BOX2D)
	CIwGameBox2dWorld*			Box2dWorld;
#endif // IW_GAME_BOX2D_H_
	CIwGameEventManager*		EventsManager;		// List of events that the scene handles
public:

	void					setParent(CIwGame* parent)				{ Parent = parent; }
	CIwGame*				getParent()								{ return Parent; }
	CIwGameSpriteManager*	getSpriteManager()						{ return SpriteManager; }
	CIwGameXomlResourceManager*	getResourceManager()				{ return ResourceManager; }
	CIwGameXomlVariableManager* getVariableManager()				{ return VariableManager; }
	void					setType(int type)						{ Type = type; }
	int						getType() const							{ return Type; }
	CIwVec2					getScreenSize() const					{ return ScreenSize; }
	CIwVec2					getVirtualSize() const					{ return VirtualSize; }
	void					setVirtualTransform(int required_width, int required_height, float angle, bool fix_aspect = false, bool lock_width = false);
	CIwMat2D&				getVirtualTransform()					{ return VirtualTransform; }
	CIwMat2D&				getTransform()							{ return Transform; }
	void					addActor(CIwGameActor *actor);
	void					removeActor(CIwGameActor* actor);
	void					removeActor(unsigned int name_hash);
	void					removeLinkedActors(CIwGameActor* actor);
	CIwGameActor*			findActor(const char* name);
	CIwGameActor*			findActor(unsigned int name_hash);
	CIwGameActor*			findActor(int type);
	CIwGameActor*			findClosestActor(int x, int y, int type);
	CIwGameActor*			findFurthestActor(int x, int y, int type);
	void					clearActors();
	void					setExtents(int x, int y, int w, int h)	{ Extents.x = x; Extents.y = y; Extents.w = w; Extents.h = h; }
	CIwRect					getExtents() const						{ return Extents; }
	void					setCamera(CIwGameCamera* camera);
	CIwGameCamera*			getCamera()								{ return Camera; }
	bool					addCollideable(CIwGameActor* actor);
	CIwGameActor**			getCollidables()						{ return Collidables; }
	int						getTotalCollidables() const				{ return NextFreeCollidable; }
	void					setAllowSuspend(bool allow_suspend)		{ AllowSuspend = allow_suspend; }
	bool					getAllowSuspend() const					{ return AllowSuspend; }
	void					setClippingArea(int x, int y, int w, int h)		{ ClippingArea.x = x; ClippingArea.y = y; ClippingArea.w = w; ClippingArea.h = h; }
	CIwRect					getClippingArea()						{ return ClippingArea; }
	void					setActive(bool active)					{ IsActive = active; }
	bool					isActive() const						{ return IsActive; }
	void					setVisible(bool visible)				{ IsVisible = visible; }
	bool					isVisible() const						{ return IsVisible; }
	void					setColour(CIwColour& colour)			{ Colour = colour; }
	CIwColour				getColour() const						{ return Colour; }
	void					setTimeline(CIwGameAnimTimeline* timeline) { Timeline = timeline; }
	CIwGameAnimTimeline*	getTimeline()							{ return Timeline; }
	void					setBindings(CIwGameXomlBindings* bindings)	{ Bindings = bindings; }
	CIwGameXomlBindings*	getBindings()							{ return Bindings; }
	void					setRunCount(int count)					{ RunCount = count; }
	int						getRunCount() const						{ return RunCount; }
	void					setDestroyed(bool destroyed)			{ Destroyed = destroyed; }
	bool					isDestroyed() const						{ return Destroyed; }
	void					setAllowFocus(bool allow)				{ AllowFocus = allow; }
	bool					getAllowFocus() const					{ return AllowFocus; }
	void					setPrevTouch(bool touch)				{ PrevTouch = touch; }
	bool					getPrevTouch() const					{ return PrevTouch; }
#if defined(IW_GAME_BOX2D)
	CIwGameBox2dWorld*		getBox2dWorld()							{ return Box2dWorld; }
#endif // IW_GAME_BOX2D_H_
	CIwGameEventManager*	getEventsManager()						{ return EventsManager; }
	// Properties end

protected:
	bool					PrevTouch;					// Used to track begin and end touch
	CIwSVec2				TransformedClipRect[2];		// Transformed clipping rect for the scene
	CIwList<CIwGameActor*>	Removals;
	int						MaxCollidables;				// Maximum allowed collidables
	int						NextFreeCollidable;			// Points to next free slot in sollidables list pool
	int						MaxLayers;					// Maximum visible layers
	virtual void			UpdateBindings();			// Update data bindings


public:
	CIwGameScene() :	IIwGameXomlResource(), Parent(NULL), Collidables(NULL), ResourceManager(NULL), NextFreeCollidable(0), Camera(NULL), MaxCollidables(0), MaxLayers(0), AllowSuspend(true), IsActive(true), 
						IsVisible(true), ClippingArea(0, 0, 0, 0), Extents(0, 0, 0, 0), RunCount(0), EventsManager(NULL), VariableManager(NULL), Bindings(NULL), Type(0), AllowFocus(false), 
						PrevTouch(false), Destroyed(false)
#if defined(IW_GAME_BOX2D)
	, Box2dWorld(NULL)
#endif // IW_GAME_BOX2D_H_
	{
		setClassType("scene");
#if defined(IW_GAME_BOX2D)
		Box2dWorld = new CIwGameBox2dWorld();
#endif // IW_GAME_BOX2D_H_
		EventsManager = new CIwGameEventManager();
	}

	virtual ~CIwGameScene();

	// After creating the scene, call Init() to initialise it, passing the maximum number of actors that you expect can collide 
	virtual int		Init(int max_collidables = 128, int max_layers = 10, bool doSleep = true);

	// Update() will update the scene and all of its contained actors
	virtual void	Update(float dt);

	// Draw() will draw all of the scenes actors
	virtual void	Draw();

	// PreDestroy() is called when the scene has been marked for deletion
	virtual void	PreDestroy() {}

	// Coordinate transform helper method
	CIwFVec2		ScreenToVirtual(float screen_x, float screen_y);		// Returns a virtual position from a screen position
	CIwVec2			VirtualToScreen(float virtual_x, float virtual_y);		// Returns a screen position frmo a virtual position

	// Event handlers (Only implement these if you need to know about these specific events)
	void			ProcessEventActions(unsigned int event_name);
	virtual void	NotifyCreate();											// This event is called when this scene is being created
	virtual void	NotifyDestroy();										// This event is called when this scene is being destroyed
	virtual void	NotifySuspending(CIwGameScene* new_scene);				// This event is called when this scene is being suspended
	virtual void	NotifyResuming(CIwGameScene* old_scene);				// This event is called when this scene is being resumed
	virtual void	NotifyLostFocus(CIwGameScene* new_scene);				// This event is called when this scene has just lost focus
	virtual void	NotifyGainedFocus(CIwGameScene* old_scene);				// This event is called when this scene has just gained focus
	virtual void	NotifyKeyBack();										// This event is called when this scene receives a back key press
	virtual void	NotifyKeyMenu();										// This event is called when this scene receives a menu key press

	// Utility 
	void			Suspend(CIwGameScene* new_scene);
	void			Resume(CIwGameScene* old_scene);
	CIwGameActor*	FindTappedActor(int x, int y);

	// Implementation of IIwGameXomlResource interface
	bool			LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node);

	// Implementation of IIwGameAnimTarget interface
	bool			UpdateFromAnimation(CIwGameAnimInstance *animation);

	// Internal
	void			RestoreClipping();
};

//
//
//
//
//  CIwGameSceneCreator - Creates an instance of a scene ovject
//
//
//
//
class CIwGameSceneCreator : public IIwGameXomlClassCreator
{
public:
	CIwGameSceneCreator()
	{
		setClassName("scene");
	}
	IIwGameXomlResource* CreateInstance(IIwGameXomlResource* parent) { return new CIwGameScene(); }
};


//
//
//
//
//  CIwGameXomlAction_Scenes - A collection of scene actions
//
//
//
//
class CIwGameXomlAction_Scenes : public IIwGameXomlAction
{
public:
	enum ActionType
	{
		Action_ChangeScene		= 0, 
		Action_SuspendScene		= 1, 
		Action_ResumeScene		= 2, 
		Action_HideScene		= 3, 
		Action_ShowScene		= 4, 
		Action_ActivateScene	= 5, 
		Action_DeactivateScene	= 6, 
		Action_KillScene		= 7, 
		Action_KillAllScenes	= 8, 
		Action_Max				= 9, 
	};
protected:
	ActionType	Type;
	CIwGameXomlAction_Scenes() {}
public:
	CIwGameXomlAction_Scenes(ActionType type)
	{
		Type = type;
		switch (Type)
		{
		case Action_ChangeScene:
			setActionName("changescene");
			break;
		case Action_SuspendScene:
			setActionName("suspendscene");
			break;
		case Action_ResumeScene:
			setActionName("resumescene");
			break;
		case Action_HideScene:
			setActionName("hidescene");
			break;
		case Action_ShowScene:
			setActionName("showscene");
			break;
		case Action_ActivateScene:
			setActionName("activatescene");
			break;
		case Action_DeactivateScene:
			setActionName("deactivatescene");
			break;
		case Action_KillScene:
			setActionName("killscene");
			break;
		case Action_KillAllScenes:
			setActionName("killallscenes");
			break;
		}
	}
	void Execute(IIwGameXomlResource* source, CIwGameAction* action);
};

#endif // IW_GAME_SCENE_H_