﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameBox2d.h"
#include "IwGameActor.h"
#include "IwGameScene.h"
#include "IwGameResources.h"

#if defined(IW_GAME_BOX2D)

// Glopbally defined collision collback used to marshal cross actor collision calls
CIwGameBox2dCcollisionCallback	g_CIwGameBox2dCcollisionCallback;

//
//
//
//
// CIwGameBox2dCcollisionCallback - Box2D collision callback
//
// These methods basically notify each actor that a collision has either started or ended
//
//
//
//
void CIwGameBox2dCcollisionCallback::BeginContact(b2Contact* contact)
{
	CIwGameBox2dBody* bodyUserDataA = (CIwGameBox2dBody*)contact->GetFixtureA()->GetBody()->GetUserData();
	CIwGameBox2dBody* bodyUserDataB = (CIwGameBox2dBody*)contact->GetFixtureB()->GetBody()->GetUserData();
	if (bodyUserDataA != NULL && bodyUserDataB != NULL)
	{
		bodyUserDataA->addCollisionStarted(bodyUserDataB);
		bodyUserDataB->addCollisionStarted(bodyUserDataA);
	}
}
 
void CIwGameBox2dCcollisionCallback::EndContact(b2Contact* contact)
{
	CIwGameBox2dBody* bodyUserDataA = (CIwGameBox2dBody*)contact->GetFixtureA()->GetBody()->GetUserData();
	CIwGameBox2dBody* bodyUserDataB = (CIwGameBox2dBody*)contact->GetFixtureB()->GetBody()->GetUserData();
	if (bodyUserDataA != NULL && bodyUserDataB != NULL)
	{
		bodyUserDataA->addCollisionEnded(bodyUserDataB);
		bodyUserDataB->addCollisionEnded(bodyUserDataA);
	}
}


//
//
//
//
// CIwGameBox2dMaterial implementation
//
//
//
//
bool CIwGameBox2dMaterial::LoadFromXoml(IIwGameXomlResource* parent, bool load_children, CIwGameXmlNode* node)
{
	// Process material attributes
	for (CIwGameXmlNode::_AttribIterator it = node->attribs_begin(); it != node->attribs_end(); it++)
	{
		unsigned int name_hash = (*it)->getName().getHash();

		if (name_hash == CIwGameXomlNames::Name_Hash)
		{
			setName((*it)->GetValue().c_str());
		}
		else
		if (name_hash == CIwGameXomlNames::Type_Hash)
		{
			unsigned int type_hash = (*it)->GetValue().getHash();
			if (type_hash == CIwGameXomlNames::static_Hash)
				setBodyType(b2_staticBody);
			else
			if (type_hash == CIwGameXomlNames::dynamic_Hash)
				setBodyType(b2_dynamicBody);
			else
			if (type_hash == CIwGameXomlNames::kinematic_Hash)
				setBodyType(b2_kinematicBody);
			else
			{
				CIwGameError::LogError("Warning: XOML - Invalid body type for box2d material - ", (*it)->GetValue().c_str());
			}
		}
		else
		if (name_hash == CIwGameXomlNames::Density_Hash)
		{
			setDensity((*it)->GetValueAsFloat());
		}
		else
		if (name_hash == CIwGameXomlNames::Friction_Hash)
		{
			setFriction((*it)->GetValueAsFloat());
		}
		else
		if (name_hash == CIwGameXomlNames::Restitution_Hash)
		{
			setRestitution((*it)->GetValueAsFloat());
		}
	}

	// If we are declared inside a scene then material is local to the scene
	if (parent != NULL && parent->getClassTypeHash() == CIwGameXomlNames::Scene_Hash)
	{
		((CIwGameScene*)parent)->getResourceManager()->addResource(this);
	}
	else
	{
		IW_GAME_GLOBAL_RESOURCES->getResourceManager()->addResource(this);
	}

	return true;
}

//
//
//
//
// CIwGameBox2dWorld implementation
//
//
//
//
CIwGameBox2dWorld::~CIwGameBox2dWorld()
{
	SAFE_DELETE(World);
}

void CIwGameBox2dWorld::setGravity(float x, float y)
{
	Gravity.x = x; Gravity.y = y;
	if (World != NULL)
		World->SetGravity(b2Vec2(Gravity.x, Gravity.y));
}

void CIwGameBox2dWorld::InitWorld(bool doSleep)
{
	// create Box2D world
	World = new b2World(b2Vec2(Gravity.x, Gravity.y), doSleep);

	// Set up contact collision listener callback (this gets called whever a collision occurs between two bodies)
	World->SetContactListener(&g_CIwGameBox2dCcollisionCallback);
}

void CIwGameBox2dWorld::UpdateWorld(float dt)
{
	if (World != NULL && WorldScale.x != 0)
	{
		// Update the physics engine
		World->Step((dt * FRAME_SPEED_LOCK_MS) / 1000.0f, VelocityIterations, PositionIterations);
	}
}



//
//
//
//
// CIwGameBox2dBody implementation
//
//
//
//

CIwGameBox2dBody::~CIwGameBox2dBody()
{
	ReleaseBody();
}

void CIwGameBox2dBody::setBody(b2Body* body)
{
	// Destroy existing body
	if (Body != NULL)
	{
		World->getWorld()->DestroyBody(Body);
		Body = NULL;
	}

	// Set new body
	Body = body;

	// Reset fixture
	Fixture = NULL;
}

void CIwGameBox2dBody::setBodyMaterial(CIwGameBox2dMaterial* mat)
{
	BodyMaterial = mat;
	if (Fixture != NULL)
	{
		Fixture->SetDensity(mat->getDensity());
		Fixture->SetFriction(mat->getFriction());
		Fixture->SetRestitution(mat->getRestitution());

		if (Body != NULL)
			Body->SetType(mat->getBodyType());
	}
}

void CIwGameBox2dBody::setFixture(b2Fixture* fixture)
{
	// Destory old fixture
	if (Fixture != NULL)
	{
		if (Body != NULL)
		{
			Body->DestroyFixture(Fixture);
		}
	}

	// Set new fixture
	Fixture = Fixture;
}

void CIwGameBox2dBody::setCollisionFlags(int category_flags, int mask_flags, int collision_group)
{
	if (Fixture != NULL)
	{
		b2Filter f;
		f.categoryBits = category_flags;
		f.maskBits = mask_flags;
		f.groupIndex = collision_group;

		Fixture->SetFilterData(f);
	}
}

int	CIwGameBox2dBody::getCollisionCategory() const
{
	if (Fixture == NULL)
		return 0;

	return Fixture->GetFilterData().categoryBits;
}

int	CIwGameBox2dBody::getCollisionMask() const
{
	if (Fixture == NULL)
		return 0;

	return Fixture->GetFilterData().maskBits;
}

int	CIwGameBox2dBody::getCollisionGroup() const
{
	if (Fixture == NULL)
		return 0;

	return Fixture->GetFilterData().groupIndex;
}

void CIwGameBox2dBody::InitBody(CIwGameBox2dWorld* world, CIwGameShape* body_shape, CIwGameBox2dMaterial* body_mat, CIwFVec2* pos, float angle, float com_x, float com_y)
{
	// Set the box2d world
	World = world;

	// Add basic fixture to the body
	b2Vec2 cm;
	cm.x = World->PixelToWorldX(com_x);
	cm.y = World->PixelToWorldX(com_y);
	b2FixtureDef fixture1;
	b2PolygonShape poly_shape;
	b2CircleShape circ_shape;

	switch (body_shape->getShape()->ShapeType)
	{
	case CIwGameGeomShape::ST_Box:
		{
			CIwGameGeomShapeBox* shp = (CIwGameGeomShapeBox*)body_shape->getShape();
			fixture1.shape = &poly_shape;
			poly_shape.SetAsBox(World->PixelToWorldX(shp->Width) / 2.0f, World->PixelToWorldY(shp->Height) / 2.0f, cm, 0);
		}
		break;
	case CIwGameGeomShape::ST_Circle:
		{
			CIwGameGeomShapeCircle* shp = (CIwGameGeomShapeCircle*)body_shape->getShape();
			fixture1.shape = &circ_shape;
			circ_shape.m_radius = World->PixelToWorldX(shp->Radius);
			circ_shape.m_p.x = cm.x;
			circ_shape.m_p.y = cm.y;
		}
		break;
	case CIwGameGeomShape::ST_Polygon:
		{
			CIwGameGeomShapePolygon* shp = (CIwGameGeomShapePolygon*)body_shape->getShape();
			fixture1.shape = &poly_shape;
			poly_shape.m_vertexCount = shp->NumVertices;
			CIwFVec2* points = shp->Vertices;

			b2Vec2 verts[8];
			for (int t = 0; t < shp->NumVertices; t++)
			{
				verts[t].x = World->PixelToWorldX(points->x) + cm.x;
				verts[t].y = World->PixelToWorldY(points->y) + cm.y;
				points++;
			}
			poly_shape.Set(verts, shp->NumVertices);
		}
		break;
	}

	b2BodyDef body_def;
	if (body_mat != NULL)
	{
		body_def.type = body_mat->getBodyType();
		body_def.position.Set(World->PixelToWorldX(pos->x), World->PixelToWorldX(pos->y));
		body_def.angle = (-angle * 3.1415927f) / 180.0f;			// Degrees to radians
		Body = World->getWorld()->CreateBody(&body_def);	
		Body->SetUserData((void*)this);
		fixture1.density = body_mat->getDensity();
		fixture1.friction = body_mat->getFriction();
		fixture1.restitution = body_mat->getRestitution();
	}
	else
	{
		body_def.type = b2_staticBody;
		body_def.position.Set(World->PixelToWorldX(pos->x), World->PixelToWorldX(pos->y));
		body_def.angle = (-angle * 3.1415927f) / 180.0f;			// Degrees to radians
		Body = World->getWorld()->CreateBody(&body_def);	
		Body->SetUserData((void*)this);
		fixture1.density = 1.0f;
		fixture1.friction = 1.0f;
		fixture1.restitution = 0.1f;
	}
	Fixture = Body->CreateFixture(&fixture1);
	BodyShape = body_shape;
	BodyMaterial = body_mat;
}

void CIwGameBox2dBody::ReleaseBody()
{
	// Destroy the body
	if (Body != NULL)
	{
		World->getWorld()->DestroyBody(Body);
		Body = NULL;
	}
}

void CIwGameBox2dBody::ApplyForce(float force_x, float force_y, float world_pos_x, float world_pos_y)
{
	if (Body != NULL)
		Body->ApplyForce(b2Vec2(force_x, force_y), b2Vec2(world_pos_x, world_pos_y));
}
void CIwGameBox2dBody::ApplyForceToCenter(float force_x, float force_y)
{
	if (Body != NULL)
		Body->ApplyForceToCenter(b2Vec2(force_x, force_y));
}
void CIwGameBox2dBody::ApplyTorque(float torque)
{
	if (Body != NULL)
		Body->ApplyTorque(torque);
}
void CIwGameBox2dBody::ApplyLinearImpulse(float impulse_x, float impulse_y, float world_pos_x, float world_pos_y)
{
	if (Body != NULL)
		Body->ApplyLinearImpulse(b2Vec2(impulse_x, impulse_y), b2Vec2(world_pos_x, world_pos_y));
}
void CIwGameBox2dBody::ApplyAngularImpulse(float impulse)
{
	if (Body != NULL)
		Body->ApplyAngularImpulse(impulse);
}

#endif	// IW_GAME_BOX2D
