// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameSprite.h"
#include "IwGameString.h"
#include "IwGameUtil.h"
#include "IwGameRender2d.h"

//
//
//
//
// CIwGameSprite implementation
//
//
//
//
void CIwGameSprite::RebuildTransform()
{
	// Build the transform
	AccumDepth = Depth;
	CIwMat2D trans;
	trans.SetIdentity();
	trans.SetTrans(Origin);
	// Set the rotation transform
	Transform.SetRot(Angle);
	// Scale the transform
	Transform.m[0][0] = IW_FIXED_MUL(Transform.m[0][0], ScaleX);
	Transform.m[0][1] = IW_FIXED_MUL(Transform.m[0][1], ScaleX);
	Transform.m[1][1] = IW_FIXED_MUL(Transform.m[1][1], ScaleY);
	Transform.m[1][0] = IW_FIXED_MUL(Transform.m[1][0], ScaleY);
	// Translate the transform
	Transform.SetTrans(Position);
	// Apply origin
	Transform.PreMult(trans);
	// Apply linked sprites transform if linked
	if (LinkedTo != NULL)
	{
		Transform.PostMult(LinkedTo->getTransform());
		AccumDepth = Depth + LinkedTo->getAccumDepth();
	}
/*	else
	{
		// Apply parent transform if sprite is managed by a parent sprite manager (doesnt work with none uniform scaling)
		if (Parent != NULL)
			Transform.PostMult(Parent->getTransform());
	}*/

	TransformDirty = false;
}

void CIwGameSprite::RebuildTransformNow()
{
	if (TransformDirty)
	{
		RebuildTransform();
		TransformVertices();
	}
}

void CIwGameSprite::Init()
{
	Parent = NULL;
	LinkedTo = NULL;
	TransformDirty = true;
	Position.x = 0;
	Position.y = 0;
	Origin.x = 0;
	Origin.y = 0;
	Angle = 0;
	ScaleX = IW_GEOM_ONE;
	ScaleY = IW_GEOM_ONE;
	Colour.r = 0xff;
	Colour.g = 0xff;
	Colour.b = 0xff;
	Colour.a = 0xff;
	Width = 0;
	Height = 0;
	Depth = IW_GEOM_ONE;
	AccumDepth = Depth;
	Visible = true;
	Layer = 0;
	FinalTransform.SetIdentity();
	Transform.SetIdentity();
}

bool CIwGameSprite::HitTest(int x, int y)
{
	x *= 8;
	y *= 8;

	int i1 = 0;
	int i2 = 3;
	for (int t = 0; t < 4; t++)
	{
		int x0 = TransformedV[i1].x - TransformedV[i2].x;
		int y0 = TransformedV[i1].y - TransformedV[i2].y;
		int x1 = x - TransformedV[i2].x;
		int y1 = y - TransformedV[i2].y;

		if ((x1 * y0 - x0 * y1) >= 0)
			return false;

		i2 = i1;
		i1++;
	}

	return true;
}

void CIwGameSprite::TransformVertices()
{
	// Apply parent transform if sprite is managed by a parent sprite manager (doesnt work with none uniform scaling)
	FinalTransform = Transform;
	if (Parent != NULL)
	{
		if (AccumDepth == IW_GEOM_ONE)
		{
			FinalTransform.PostMult(Parent->getTransform());
		}
		else
		{
			CIwSVec2 cop = Parent->getCOP();
			CIwSVec2 centre = Parent->getScreenCentre();

			CIwMat2D m = Parent->getTransform();
			m.m[0][0] = (m.m[0][0] << IW_GEOM_POINT) / AccumDepth;
			m.m[0][1] = (m.m[0][1] << IW_GEOM_POINT) / AccumDepth;
			m.m[1][0] = (m.m[1][0] << IW_GEOM_POINT) / AccumDepth;
			m.m[1][1] = (m.m[1][1] << IW_GEOM_POINT) / AccumDepth;
			m.t.x = ((m.t.x - centre.x) << IW_GEOM_POINT) / AccumDepth + cop.x;
			m.t.y = ((m.t.y - centre.y) << IW_GEOM_POINT) / AccumDepth + cop.y;

			FinalTransform.PostMult(m);
		}
	}
	// Remove sub-pixel precision (causes nasty artifacts)
//	FinalTransform.t.x &= 0xfffffff8;
//	FinalTransform.t.y &= 0xfffffff8;
//	FinalTransform.t.x = (FinalTransform.t.x << IW_GEOM_POINT) / AccumDepth;
//	FinalTransform.t.y = (FinalTransform.t.y << IW_GEOM_POINT) / AccumDepth;

	// Generate transformed vertices
	int x0 = -(Width * 4) - 4;
	int y0 = -(Height * 4) - 4;
	int x1 = (Width * 4) + 4;
	int y1 = (Height * 4) + 4;
	TransformedV[0].x = x0;
	TransformedV[1].x = x1;
	TransformedV[2].x = x1;
	TransformedV[3].x = x0;
	TransformedV[0].y = y0;
	TransformedV[1].y = y0;
	TransformedV[2].y = y1;
	TransformedV[3].y = y1;
	for (int t = 0; t < 4; t++)
		TransformedV[t] = FinalTransform.TransformVec(TransformedV[t]);
}

bool CIwGameSprite::Update()
{
	// Do not render if not visible
	if (!Visible || Colour.a == 0)
		return false;

	if (LinkedTo != NULL)
	{
		if (!LinkedTo->isVisible())
			return false;
	}

	if (TransformDirty)
	{
		RebuildTransform();
		TransformVertices();
	}

	return true;
}

//
//
//
//
// CIwGameDummySprite implementation
//
//
//
//

void CIwGameDummySprite::Draw()
{
}


//
//
//
//
// CIwGameBitmapSprite implementation
//
//
//
//
void CIwGameBitmapSprite::RebuildUVList()
{
	if (Image == NULL)
		return;

	int		w = Image->getWidth();
	int		h = Image->getHeight();
	int		px = ((IW_GEOM_ONE / w) >> 1);
	int		py = ((IW_GEOM_ONE / h) >> 1);
//	int		px = 1;
//	int		py = 1;
	int		x0 = (SrcX * IW_GEOM_ONE) / w, x1 = ((SrcX + SrcWidth) * IW_GEOM_ONE) / w - px;
	int		y0 = (SrcY * IW_GEOM_ONE) / h, y1 = ((SrcY + SrcHeight) * IW_GEOM_ONE) / h - py;
//	int		x0 = (SrcX * IW_GEOM_ONE) / w, x1 = ((SrcX + SrcWidth) * IW_GEOM_ONE) / w;
//	int		y0 = (SrcY * IW_GEOM_ONE) / h, y1 = ((SrcY + SrcHeight) * IW_GEOM_ONE) / h;

	if (ImageTransform & IW2D_FLIP_X_F)
	{
		int t = x0;
		x0 = x1;
		x1 = t;
	}
	if (ImageTransform & IW2D_FLIP_Y_F)
	{
		int t = y0;
		y0 = y1;
		y1 = t;
	}

	UVList[0].x = x0;
	UVList[0].y = y0;
	UVList[1].x = x1;
	UVList[1].y = y0;
	UVList[2].x = x1;
	UVList[2].y = y1;
	UVList[3].x = x0;
	UVList[3].y = y1;
	UVsDirty = false;
}

void CIwGameBitmapSprite::Draw()
{
	// Do not render if not visible
	if (Image == NULL || Image->getState() != CIwGameImage::CIwGameImage_State_Loaded)
		return;

	if (UVsDirty)
		RebuildUVList();

	// Set image transform
//	Iw2DSetImageTransform(ImageTransform);

	// Set this transform as the active transform for Iw2D
//	Iw2DSetTransformMatrixSubPixel(Transform);

	// Set colour of sprite
//	Iw2DSetColour(Colour);


	// Render the sprite (centered)
//	int x = -(Width / 2) * 8;
//	int y = -(Height / 2) * 8;
	if (Parent->getBatching())
		IW_GAME_RENDER2D->AddPrimtive(TransformedV, UVList, Image, Colours);
	else
		IW_GAME_RENDER2D->DrawImage(TransformedV, UVList, Image, Colours);
//	if (Image->getState() == CIwGameImage::CIwGameImage_State_Loaded)
//		Iw2DDrawImageRegion(Image->getImage2D(), CIwSVec2(x, y), CIwSVec2(Width * 8, Height * 8), CIwSVec2(SrcX, SrcY), CIwSVec2(SrcWidth, SrcHeight));
//	Iw2DDrawImage(Image, CIwSVec2(x, y));
}

//
//
//
//
// CIwGameTextSprite implementation
//
//
//
//
void CIwGameTextSprite::RebuildTransform()
{
	// Build the transform
	AccumDepth = Depth;
	CIwMat2D trans;
	trans.SetIdentity();
	trans.SetTrans(Origin);
	trans.t.x -= Rect.w << 2;
	trans.t.y -= Rect.h << 2;
	// Set the rotation transform
	Transform.SetRot(Angle);
	// Scale the transform
	Transform.m[0][0] = IW_FIXED_MUL(Transform.m[0][0], ScaleX);
	Transform.m[0][1] = IW_FIXED_MUL(Transform.m[0][1], ScaleX);
	Transform.m[1][1] = IW_FIXED_MUL(Transform.m[1][1], ScaleY);
	Transform.m[1][0] = IW_FIXED_MUL(Transform.m[1][0], ScaleY);
	// Translate the transform
	Transform.SetTrans(Position);
	// Apply origin
	Transform.PreMult(trans);
	// Apply linked sprites transform if linked
	if (LinkedTo != NULL)
	{
		Transform.PostMult(LinkedTo->getTransform());
		AccumDepth = Depth + LinkedTo->getAccumDepth();
	}
/*	else
	{
		// Apply parent transform if sprite is managed by a parent sprite manager (doesnt work with none uniform scaling)
		if (Parent != NULL)
			Transform.PostMult(Parent->getTransform());
	}*/

	TransformDirty = false;
}

void CIwGameTextSprite::TransformVertices()
{
	// Apply parent transform if sprite is managed by a parent sprite manager (doesnt work with none uniform scaling)
	FinalTransform = Transform;
	if (Parent != NULL)
	{
		if (AccumDepth == IW_GEOM_ONE)
		{
			FinalTransform.PostMult(Parent->getTransform());
		}
		else
		{
			CIwSVec2 cop = Parent->getCOP();
			CIwSVec2 centre = Parent->getScreenCentre();

			CIwMat2D m = Parent->getTransform();
			m.m[0][0] = (m.m[0][0] << IW_GEOM_POINT) / AccumDepth;
			m.m[0][1] = (m.m[0][1] << IW_GEOM_POINT) / AccumDepth;
			m.m[1][0] = (m.m[1][0] << IW_GEOM_POINT) / AccumDepth;
			m.m[1][1] = (m.m[1][1] << IW_GEOM_POINT) / AccumDepth;
			m.t.x = ((m.t.x - centre.x) << IW_GEOM_POINT) / AccumDepth + cop.x;
			m.t.y = ((m.t.y - centre.y) << IW_GEOM_POINT) / AccumDepth + cop.y;

			FinalTransform.PostMult(m);
		}
	}

	// Generate transformed vertices
	int x0 = -4;
	int y0 = -4;
	int x1 = (Rect.w * 8) + 4;
	int y1 = (Rect.h * 8) + 4;
	TransformedV[0].x = x0;
	TransformedV[1].x = x1;
	TransformedV[2].x = x1;
	TransformedV[3].x = x0;
	TransformedV[0].y = y0;
	TransformedV[1].y = y0;
	TransformedV[2].y = y1;
	TransformedV[3].y = y1;
	for (int t = 0; t < 4; t++)
		TransformedV[t] = FinalTransform.TransformVec(TransformedV[t]);
}

void CIwGameTextSprite::RebuildText()
{
	if (Font == NULL)
		return;

	IwGxFontSetScale(IW_GEOM_ONE);
    IwGxFontPrepareText(PreparedText, Text.c_str());

	TextDirty = false;
}

void CIwGameTextSprite::Draw()
{
	// Do not render if not visible
	if (Font == NULL)
		return;

	IwGxFontSetFont(Font->getFont());
	IwGxFontSetFlags(Flags);
	IwGxFontSetRect(Rect);
	IwGxFontSetAlignmentHor(AlignH);
	IwGxFontSetAlignmentVer(AlignV);

	if (TextDirty)
		RebuildText();

	// Cannot batch as we cannot get access to glyph UV coordinates
	IW_GAME_RENDER2D->DrawText(&PreparedText, Font, &FinalTransform, Colour);
}

//
//
//
// CIwGameSpriteManager implementation
//
//
//
void CIwGameSpriteManager::addSprite(CIwGameSprite* sprite)
{
	Sprites.push_back(sprite);
	sprite->setParent(this);
}

void CIwGameSpriteManager::removeSprite(CIwGameSprite* sprite, bool delete_sprites)
{
	for (Iterator it = Sprites.begin(); it != Sprites.end(); ++it)
	{
		if (*it == sprite)
		{
			if (delete_sprites && !(*it)->isPooled())
				delete *it;
			Sprites.erase(it);
			break;
		}
	}
}

void CIwGameSpriteManager::Init(int max_layers)
{
	// Allocate layers
	Layers = new CIwList<CIwGameSprite*>[max_layers];
	MaxLayers = max_layers;
}

void CIwGameSpriteManager::Draw()
{
//	bool sub_pixel = Iw2DGetSubPixelCoords();

	// Enable sub pixel coordinates
//	Iw2DSetSubPixelCoords(true);

	// Clear layers
	ClearLayers();

	// Update and organise sprites into layers
	for (Iterator it = Sprites.begin(); it != Sprites.end(); ++it)
	{
		if ((*it)->Update())
		{
			int layer = (*it)->getLayer();
			(Layers + layer)->push_back(*it);
		}
	}

	// Draw sprite layers
	CIwList<CIwGameSprite*>* layers = Layers;
	for (int t = 0; t < MaxLayers; t++)
	{
		for (Iterator it = layers->begin(); it != layers->end(); ++it)
		{
			(*it)->Draw();
		}
		if (Batching)
			IW_GAME_RENDER2D->BatchDrawPrims();
		layers++;
	}

//	Iw2DSetSubPixelCoords(sub_pixel);
}

void CIwGameSpriteManager::Release(bool delete_sprites)
{
	if (delete_sprites)
	{
		// Delete all sprites in the sprite manager
		for (Iterator it = Sprites.begin(); it != Sprites.end(); ++it)
		{
			if (!(*it)->isPooled())
				delete *it;
		}
	}
	Sprites.clear();

	// Clean up layers
	SAFE_DELETE_ARRAY(Layers)
}

void CIwGameSpriteManager::DirtyChildTransforms()
{
	// Force all childreen to rebuild their transforms
	for (Iterator it = Sprites.begin(); it != Sprites.end(); ++it)
	{
		(*it)->forceTransformDirty();
	}
}

void CIwGameSpriteManager::ClearLayers()
{
	for (int t = 0; t < MaxLayers; t++)
	{
		(Layers + t)->clear();
	}
}

